"""Defines a task mixin for profiling PyTorch models."""

import contextlib
import datetime
import logging
import time
from dataclasses import dataclass
from typing import Any, ContextManager, Generic, Iterator, TypeVar

import torch

from mlfab.core.conf import field
from mlfab.nn.device.gpu import gpu_device
from mlfab.task.mixins.artifacts import ArtifactsConfig, ArtifactsMixin
from mlfab.task.mixins.logger import LoggerConfig, LoggerMixin
from mlfab.task.mixins.step_wrapper import StepContextConfig, StepContextMixin, StepType

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class ProfilerOptions:
    enabled: bool = field(False, help="If profiling should be enabled")
    record_shapes: bool = field(False, help="If set, record tensor shapes")
    profile_memory: bool = field(False, help="If set, profile PyTorch memory")
    with_stack: bool = field(False, help="Record source information (file and line number) for ops")
    with_flops: bool = field(False, help="Use formula to estimate the FLOPs of specific operations")
    with_modules: bool = field(False, help="Record module hierarchy (including function names)")
    wait_steps: int = field(10, help="Number of initial waiting steps")
    warmup_steps: int = field(10, help="Number of profiler warmup steps")
    active_steps: int = field(10, help="Number of profiler active steps")
    repeat_steps: int = field(1, help="Number of profiler repetitions")
    skip_first_steps: int = field(10, help="Number of profiler steps to skip at first")
    table_size: int = field(10, help="Number of profiling ops to print")


STEPS_TO_TIME: set[StepType] = {
    "backward",
    "clip_grads",
    "forward",
    "get_single_loss",
    "log_losses",
    "on_step_end",
    "on_step_start",
    "step",
    "write_logs",
    "zero_grads",
}


STEPS_TO_WARN_IF_LONG: set[StepType] = {
    "create_optimizers",
    "get_dataloader",
    "get_dataset",
    "get_prefetcher",
    "load_checkpoint",
    "model_to_device",
    "save_checkpoint",
}


@dataclass
class ProfilerConfig(LoggerConfig, StepContextConfig, ArtifactsConfig):
    profiler: ProfilerOptions = field(ProfilerOptions(), help="Profiler configuration")


Config = TypeVar("Config", bound=ProfilerConfig)


class ProfilerMixin(
    LoggerMixin[Config],
    StepContextMixin[Config],
    ArtifactsMixin[Config],
    Generic[Config],
):
    """Defines a task mixin for enabling the PyTorch profiler."""

    def warn_if_step_too_long(self, step: StepType, seconds: float) -> bool:
        return seconds > 5.0

    def step_context(self, step: StepType) -> ContextManager:
        ctx = super().step_context(step)

        if step in STEPS_TO_TIME:

            @contextlib.contextmanager
            def wrapped_timer_ctx() -> Iterator[Any]:
                start_time = time.time()

                if self.config.profiler.enabled:
                    with ctx as a, torch.profiler.record_function(step) as b:
                        yield a, b
                else:
                    with ctx as a:
                        yield a

                step_time = time.time() - start_time
                self.log_scalar(step, step_time, namespace="🔧 dt")

            return wrapped_timer_ctx()

        if step in STEPS_TO_WARN_IF_LONG:

            @contextlib.contextmanager
            def wrapped_warn_if_long_ctx() -> Iterator[Any]:
                start_time = time.time()

                if self.config.profiler.enabled:
                    with ctx as a, torch.profiler.record_function(step) as b:
                        yield a, b
                else:
                    with ctx as a:
                        yield a

                step_time = time.time() - start_time
                if self.warn_if_step_too_long(step, step_time):
                    logger.warning("Step %s took %.2f seconds", step, step_time)

            return wrapped_warn_if_long_ctx()

        return ctx

    def on_profiler_trace_ready(self, prof: torch.profiler.profile) -> None:
        key_averages = prof.key_averages()

        # Prints a table with informative statistics.
        keys = ["self_cpu_time_total", "cpu_time_total", "cpu_memory_usage"]
        if gpu_device.has_device():
            keys += ["self_cuda_time_total", "cuda_time_total", "cuda_memory_usage"]
        for key in keys:
            table = key_averages.table(
                sort_by=key,
                row_limit=self.config.profiler.table_size,
                top_level_events_only=False,
            )
            logger.info("%s:\n%s", key, table)

        # Saves a stack trace that is viewable in Chrome, in chrome://tracing/
        profile_dir = self.exp_dir / "profile"
        profile_dir.mkdir(exist_ok=True, parents=True)
        date_str = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        prof.export_chrome_trace(str(profile_dir / f"trace.step_{prof.step_num}.{date_str}.json"))

    def get_profile(self) -> torch.profiler.profile | None:
        if not self.config.profiler.enabled:
            return None

        if gpu_device.has_device():
            profiler_activities = [
                torch.autograd.ProfilerActivity.CPU,
                torch.autograd.ProfilerActivity.CUDA,
            ]
        else:
            profiler_activities = [
                torch.autograd.ProfilerActivity.CPU,
            ]

        return torch.profiler.profile(
            activities=profiler_activities,
            record_shapes=self.config.profiler.record_shapes,
            profile_memory=self.config.profiler.profile_memory,
            with_stack=self.config.profiler.with_stack,
            with_flops=self.config.profiler.with_flops,
            with_modules=self.config.profiler.with_modules,
            schedule=torch.profiler.schedule(
                wait=self.config.profiler.wait_steps,
                warmup=self.config.profiler.warmup_steps,
                active=self.config.profiler.active_steps,
                repeat=self.config.profiler.repeat_steps,
                skip_first=self.config.profiler.skip_first_steps,
            ),
            on_trace_ready=self.on_profiler_trace_ready,
        )



# mypy: disable-error-code="import, no-untyped-def, override"
# ruff: noqa: ANN001, ANN201, ANN202, N803, N806
"""Defines Triton kernels for the log-space RWKV forward and backward passes."""

import warnings
from typing import Any

import torch
import triton
import triton.language as tl
from torch import Tensor
from torch.autograd.function import Function, FunctionCtx, once_differentiable

MIN_LOG_PROB = -1e4


def get_block_size_c(chans: int) -> int:
    if chans < 32:
        return 32
    if chans >= 2048:
        # As written, the kernels need to be able to pass values between
        # adjacent threads, meaning that we need to be able to keep the entire
        # sequence in shared memory. We set an upper bound of 2048, which is
        # probably longer than any real-world sequence we might care about, and
        # limit ourselves to at most one block.
        raise NotImplementedError("Triton kernels do not support more than 2048 target sequence length")
    return triton.next_power_of_2(chans)


@triton.jit
def logaddexp(a, b):
    max_ab = tl.maximum(a, b)
    return max_ab + tl.math.log(tl.math.exp(a - max_ab) + tl.math.exp(b - max_ab))


@triton.jit
def pos_log_prob(x):
    return -tl.math.log1p(tl.math.exp(-x))


@triton.jit
def neg_log_prob(x):
    return -tl.math.log1p(tl.math.exp(x))


@triton.jit
def d_pos_log_prob(x):
    return 1 / (1 + tl.math.exp(x))


@triton.jit
def d_neg_log_prob(x):
    return -1 / (1 + tl.math.exp(-x))


def _neg_log_prob(x: Tensor) -> Tensor:
    return -torch.log1p(torch.exp(x))


def _d_neg_log_prob(x: Tensor) -> Tensor:
    return -1 / (1 + torch.exp(-x))


@triton.jit
def forward_pass_kernel(
    # Log probabilities tensor (input)
    logits_ptr,
    logits_s_bsz,
    logits_s_src,
    logits_s_tgt,
    # Log phis tensor (output)
    phis_ptr,
    phis_s_bsz,
    phis_s_src,
    phis_s_tgt,
    # Tensor dimensions
    t_i,
    t_j,
    BLOCK_SIZE_C: tl.constexpr,
):
    # Parallelize over the batch dimension.
    b_idx = tl.program_id(0)
    j_idx = tl.program_id(1)

    # Pointers to the log probabilities.
    j = (j_idx * BLOCK_SIZE_C) + tl.arange(0, BLOCK_SIZE_C)
    jmask = j < t_j
    jmask_shifted = jmask & (j > 0)

    # Gets pointers offset for the current batch.
    logits_ptr = logits_ptr + b_idx * logits_s_bsz
    phis_ptr = phis_ptr + b_idx * phis_s_bsz

    # Accumulator for the log phis.
    phis_acc = tl.where(j == 0, 0.0, tl.full((BLOCK_SIZE_C,), value=MIN_LOG_PROB, dtype=tl.float32))

    # Stores first log phi value.
    phis_first_ptr = phis_ptr + j * phis_s_tgt
    tl.store(phis_first_ptr, phis_acc, mask=jmask)
    tl.debug_barrier()

    for i in range(1, t_i):
        logits_prev_ptr = logits_ptr + (i - 1) * logits_s_src + j * logits_s_tgt
        logits_prev = tl.load(logits_prev_ptr, mask=jmask).to(tl.float32)

        phis_prev_m1_ptr = phis_ptr + (i - 1) * phis_s_src + (j - 1) * phis_s_tgt
        logits_prev_m1_ptr = logits_ptr + (i - 1) * logits_s_src + (j - 1) * logits_s_tgt
        phis_prev_m1 = tl.load(phis_prev_m1_ptr, mask=jmask_shifted, other=MIN_LOG_PROB).to(tl.float32)
        logits_prev_m1 = tl.load(logits_prev_m1_ptr, mask=jmask_shifted, other=MIN_LOG_PROB).to(tl.float32)

        phis_a = phis_prev_m1 + neg_log_prob(logits_prev_m1)
        phis_b = phis_acc + pos_log_prob(logits_prev)
        phis_acc = logaddexp(phis_a, phis_b)

        phis_next_ptr = phis_ptr + i * phis_s_src + j * phis_s_tgt
        tl.store(phis_next_ptr, phis_acc, mask=jmask)

        # Barrier to ensure that we can access the stored log phis from the
        # adjacent thread in the next iteration.
        tl.debug_barrier()


@triton.jit
def backward_pass_kernel(
    # Log probabilities tensor (input)
    logits_ptr,
    logits_stride_bsz,
    logits_stride_src,
    logits_stride_tgt,
    # Log phis tensor (input)
    phis_ptr,
    phis_s_bsz,
    phis_s_src,
    phis_s_tgt,
    # Gradient of log phis tensor (input)
    grad_phis_ptr,
    grad_phis_s_bsz,
    grad_phis_s_src,
    grad_phis_s_tgt,
    # Gradient of log probabilities tensor (output)
    grad_logits_ptr,
    grad_logits_s_bsz,
    grad_logits_s_src,
    grad_logits_s_tgt,
    # Tensor dimensions
    t_i,
    t_j,
    BLOCK_SIZE_C: tl.constexpr,
):
    # Parallelize over the batch dimension.
    b_idx = tl.program_id(0)
    j_idx = tl.program_id(1)

    # Pointers to the log probabilities.
    j = (j_idx * BLOCK_SIZE_C) + tl.arange(0, BLOCK_SIZE_C)
    jmask = j < t_j
    jmask_shifted = j < (t_j - 1)

    # Gets pointers offset for the current batch.
    logits_ptr = logits_ptr + b_idx * logits_stride_bsz
    phis_ptr = phis_ptr + b_idx * phis_s_bsz
    grad_phis_ptr = grad_phis_ptr + b_idx * grad_phis_s_bsz
    grad_logits_ptr = grad_logits_ptr + b_idx * grad_logits_s_bsz

    # Stores first log phi value.
    grad_logits_last_ptr = grad_logits_ptr + (t_i - 1) * phis_s_src + j * phis_s_tgt
    tl.store(grad_logits_last_ptr, tl.zeros((BLOCK_SIZE_C,), dtype=tl.float32), mask=j < t_j)
    tl.debug_barrier()

    for i in range(t_i - 2, -1, -1):
        # phis[..., i + 1, :]
        phis_next_ptr = phis_ptr + (i + 1) * phis_s_src + j * phis_s_tgt
        phis_next = tl.load(phis_next_ptr, mask=jmask)

        # phis[..., i + 1, 1:]
        phis_next_p1_ptr = phis_ptr + (i + 1) * phis_s_src + (j + 1) * phis_s_tgt
        phis_next_p1 = tl.load(phis_next_p1_ptr, mask=jmask_shifted, other=0.0)

        # phis[..., i, :]
        phis_cur_ptr = phis_ptr + i * phis_s_src + j * phis_s_tgt
        phis_cur = tl.load(phis_cur_ptr, mask=jmask)

        # logits[..., i, :]
        logits_cur_ptr = logits_ptr + i * logits_stride_src + j * logits_stride_tgt
        logits_cur = tl.load(logits_cur_ptr, mask=jmask).to(tl.float32)

        # grad_phis[..., i + 1, :]
        grad_phis_next_ptr = grad_phis_ptr + (i + 1) * grad_phis_s_src + j * grad_phis_s_tgt
        grad_phis_next = tl.load(grad_phis_next_ptr, mask=jmask)

        # grad_phis[..., i + 1, 1:]
        grad_phis_next_p1_ptr = grad_phis_ptr + (i + 1) * grad_phis_s_src + (j + 1) * grad_phis_s_tgt
        grad_phis_next_p1 = tl.load(grad_phis_next_p1_ptr, mask=jmask_shifted, other=MIN_LOG_PROB)

        # grad_logits[..., i, :]
        grad_logits_cur_ptr = grad_logits_ptr + i * grad_logits_s_src + j * grad_logits_s_tgt

        # grad_phis[..., i, :]
        grad_phis_cur_ptr = grad_phis_ptr + i * grad_phis_s_src + j * grad_phis_s_tgt
        grad_phis_cur = tl.load(grad_phis_cur_ptr, mask=jmask)

        # Computes the new values.
        a = tl.math.exp(phis_cur + pos_log_prob(logits_cur) - phis_next)
        b = tl.math.exp(phis_cur + neg_log_prob(logits_cur) - phis_next_p1)
        c = grad_phis_next * a
        d = grad_phis_next_p1 * b
        grad_logits_cur = tl.where(
            jmask_shifted,
            c * d_pos_log_prob(logits_cur) + d * d_neg_log_prob(logits_cur),
            c * d_pos_log_prob(logits_cur),
        )
        grad_phis_cur = grad_phis_cur + tl.where(jmask_shifted, c + d, c)

        # Stores the new values.
        tl.store(grad_logits_cur_ptr, grad_logits_cur, mask=jmask)
        tl.store(grad_phis_cur_ptr, grad_phis_cur, mask=jmask)

        # Barrier to ensure that we can access the stored log phis from the
        # adjacent thread in the next iteration.
        tl.debug_barrier()


def forward_pass_gpu_(logits: Tensor) -> tuple[Tensor, Tensor]:
    bsz, tsz_src, tsz_tgt = logits.shape

    # Sets the initial log phi values.
    phis = torch.empty_like(logits)

    block_size_c = get_block_size_c(tsz_src)

    def grid(meta: dict[str, Any]) -> tuple[int, ...]:
        return (bsz, triton.cdiv(tsz_src, meta["BLOCK_SIZE_C"]))

    forward_pass_kernel[grid](
        # Log probabilities
        logits,
        logits.stride(0),
        logits.stride(1),
        logits.stride(2),
        # Log phis
        phis,
        phis.stride(0),
        phis.stride(1),
        phis.stride(2),
        # Tensor dimensions
        tsz_src,
        tsz_tgt,
        # Block size
        BLOCK_SIZE_C=block_size_c,
    )

    return phis, phis + _neg_log_prob(logits)


def backward_pass_gpu_(logits: Tensor, phis: Tensor, grad_phis: Tensor) -> Tensor:
    bsz, tsz_src, tsz_tgt = logits.shape

    extra_grad = _d_neg_log_prob(logits) * grad_phis
    grad_logits = torch.full_like(grad_phis, MIN_LOG_PROB)

    # We need to duplicate the phis tensor because the kernel updates it.
    grad_phis = grad_phis.clone()

    block_size_c = get_block_size_c(tsz_src)

    def grid(meta: dict[str, Any]) -> tuple[int, ...]:
        return (bsz, triton.cdiv(tsz_src, meta["BLOCK_SIZE_C"]))

    backward_pass_kernel[grid](
        # Log probabilities
        logits,
        logits.stride(0),
        logits.stride(1),
        logits.stride(2),
        # Log phis
        phis,
        phis.stride(0),
        phis.stride(1),
        phis.stride(2),
        # Gradient of log phis
        grad_phis,
        grad_phis.stride(0),
        grad_phis.stride(1),
        grad_phis.stride(2),
        # Gradient of log probabilities
        grad_logits,
        grad_logits.stride(0),
        grad_logits.stride(1),
        grad_logits.stride(2),
        # Tensor dimensions
        tsz_src,
        tsz_tgt,
        # Block size
        BLOCK_SIZE_C=block_size_c,
    )

    return grad_logits + extra_grad


class MonotonicAttentionGpu(Function):
    @staticmethod
    def forward(ctx: FunctionCtx, logits: Tensor) -> Tensor:
        phis, probs = forward_pass_gpu_(logits)
        ctx.save_for_backward(logits, phis)
        return probs

    @staticmethod
    @once_differentiable
    def backward(ctx: FunctionCtx, grad_phis: Tensor) -> Tensor:
        logits, phis = ctx.saved_tensors
        grad_logits = backward_pass_gpu_(logits, phis, grad_phis)
        return grad_logits


def monotonic_attention_gpu(logits: Tensor) -> Tensor:
    """Computes the monotonic attention normalization on the transition probabilities.

    Args:
        logits: The transition logits, with shape ``(bsz, tsz_src, tsz_tgt)``

    Returns:
        The marginalized log probabilities for each cell being part of a
        monotonic alignment path, with shape ``(bsz, tsz_src, tsz_tgt)``.
    """
    _, tsz_src, tsz_tgt = logits.size()
    if tsz_tgt > tsz_src:
        warnings.warn("One-to-many attention expects the source sequence to be longer than the target sequence!")
    return MonotonicAttentionGpu.apply(logits)



"""Defines custom dataset functions."""

from typing import Generic, Iterator, Sized, TypeVar

from torch.utils.data.dataset import Dataset, IterableDataset

T = TypeVar("T")


class SmallDataset(IterableDataset[T], Generic[T]):
    """Defines a dataset which caches a small number of items.

    This is useful for validating a model on a small dataset before running
    a larger training job. It automatically detects iterable verses
    non-iterable datasets. Note that all the elements of the dataset must fit
    into memory.

    After this dataset has filled the cache, it loops through the cached items
    until the training run is over.

    Parameters:
        size: The number of unique samples to yield from the dataset.
        dataset: The dataset to wrap.
    """

    def __init__(self, size: int, dataset: Dataset[T]) -> None:
        super().__init__()

        assert size > 0, "Size must be positive."

        self.size = self._get_size(size, dataset)
        self.counter = 0
        self.dataset = dataset
        self.cache: list[T] = []

        self._ds_iter: Iterator[T] | None = None

    def _get_size(self, max_size: int, dataset: Dataset[T]) -> int:
        if isinstance(dataset, IterableDataset):
            return max_size
        if isinstance(dataset, Dataset) and isinstance(dataset, Sized):
            return min(max_size, len(dataset))
        raise TypeError("Dataset must be iterable or indexable.")

    def _add_item(self, recursed: bool = False) -> None:
        if isinstance(self.dataset, IterableDataset):
            if self._ds_iter is None:
                self._ds_iter = iter(self.dataset)
            try:
                self.cache.append(next(self._ds_iter))
            except StopIteration:
                if recursed:
                    raise RuntimeError("Dataset doesn't yield any items!")
                self._ds_iter = None
                self._add_item(recursed=True)
        elif isinstance(self.dataset, Dataset) and isinstance(self.dataset, Sized):
            self.cache.append(self.dataset[len(self.cache)])
        else:
            raise TypeError("Dataset must be iterable or indexable.")

    def __iter__(self) -> Iterator[T]:
        self.counter = 0
        return self

    def __next__(self) -> T:
        if self.counter >= self.size:
            self.counter = 0
        while len(self.cache) <= self.counter:
            self._add_item()
        self.counter += 1
        return self.cache[self.counter - 1]

    def __len__(self) -> int:
        return self.size

    def __getitem__(self, index: int) -> T:
        assert 0 <= index < self.size, "Index out of bounds."
        while index >= len(self.cache):
            self._add_item()
        return self.cache[index]



"""Defines a mixin for handling model checkpointing."""

import json
import logging
import warnings
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Callable, Generic, Self, TypeVar, cast

import torch
from omegaconf import DictConfig, OmegaConf
from torch.serialization import MAP_LOCATION

from mlfab.core.conf import field
from mlfab.core.state import State
from mlfab.nn.parallel import is_master
from mlfab.task.mixins.artifacts import ArtifactsConfig, ArtifactsMixin
from mlfab.utils.experiments import diff_configs, get_diff_string

logger = logging.getLogger(__name__)


def get_ckpt_path(exp_dir: Path, state: State | None = None) -> Path:
    """Defines the path to the checkpoint for a given state.

    Args:
        exp_dir: The experiment directory
        state: The current trainer state

    Returns:
        The path to the PyTorch checkpoint to save or load
    """
    if state is None:
        return exp_dir / "checkpoints" / "ckpt.pt"
    return exp_dir / "checkpoints" / f"ckpt.{state.num_steps}.pt"


@dataclass
class CheckpointingConfig(ArtifactsConfig):
    save_every_n_steps: int | None = field(None, help="Save a checkpoint every N steps")
    save_every_n_seconds: float | None = field(60.0 * 60.0, help="Save a checkpoint every N seconds")
    only_save_most_recent: bool = field(True, help="Only keep the most recent checkpoint")
    load_from_ckpt_path: str | None = field(None, help="If set, load initial model weights from this path")
    load_ckpt_strict: bool = field(True, help="If set, only load weights for which have a matching key in the model")


Config = TypeVar("Config", bound=CheckpointingConfig)


class CheckpointingMixin(ArtifactsMixin[Config], Generic[Config]):
    def __init__(self, config: Config) -> None:
        super().__init__(config)

        self.__last_ckpt_time = 0.0

    def get_ckpt_path(self, state: State | None = None) -> Path:
        return get_ckpt_path(self.exp_dir, state)

    @classmethod
    def read_state_dict(
        cls,
        path: str | Path,
        map_location: MAP_LOCATION = None,
        weights_only: bool = False,
        mmap: bool | None = None,
    ) -> dict:
        return torch.load(
            path,
            map_location=map_location,
            weights_only=weights_only,
            mmap=mmap,
        )

    @classmethod
    def get_task_from_ckpt(
        cls,
        path: str | Path,
        strict: bool = True,
        assign: bool = False,
        use_cli: bool | list[str] = False,
        map_location: MAP_LOCATION = None,
        weights_only: bool = False,
        mmap: bool | None = None,
        config_fn: Callable[[DictConfig], DictConfig] = lambda x: x,
    ) -> Self:
        state_dict = cls.read_state_dict(
            path,
            map_location=map_location,
            weights_only=weights_only,
            mmap=mmap,
        )
        raw_config = state_dict.pop("config", None)
        if raw_config is None:
            raise RuntimeError(f"Could not find config in checkpoint at {path}!")
        cfg = cls.get_config(config_fn(OmegaConf.create(raw_config)), use_cli=use_cli)
        task = cls(cfg)
        task.load_task_state_dict(state_dict, strict=strict, assign=assign)
        return task

    def get_init_ckpt_path(self) -> Path | None:
        if self._exp_dir is not None:
            ckpt_path = self.get_ckpt_path()
            if ckpt_path.exists():
                return ckpt_path
        if self.config.load_from_ckpt_path is not None:
            ckpt_path = Path(self.config.load_from_ckpt_path)
            assert ckpt_path.exists(), f"Checkpoint path {ckpt_path} does not exist."
            return ckpt_path
        return None

    def load_initial_state(
        self,
        map_location: MAP_LOCATION = None,
        weights_only: bool = False,
        mmap: bool | None = None,
        strict: bool = True,
        assign: bool = False,
    ) -> State:
        init_ckpt_path = self.get_init_ckpt_path()
        if init_ckpt_path is None:
            return State.init_state()
        state_dict = self.read_state_dict(
            init_ckpt_path,
            map_location=map_location,
            weights_only=weights_only,
            mmap=mmap,
        )
        raw_state = state_dict.pop("state", None)
        raw_config = state_dict.pop("config", None)
        if raw_config is not None:
            config_diff = get_diff_string(diff_configs(OmegaConf.create(raw_config), cast(DictConfig, self.config)))
            if config_diff:
                logger.warning("Loaded config differs from current config:\n%s", config_diff)
        self.load_task_state_dict(state_dict, strict, assign)
        if raw_state is not None:
            return State(**json.loads(raw_state))
        warnings.warn("No state found in checkpoint! Using default initial state.")
        return State.init_state()

    def should_checkpoint(self, state: State) -> bool:
        if self.config.save_every_n_steps is not None:
            if state.num_steps % self.config.save_every_n_steps == 0:
                return True
        if self.config.save_every_n_seconds is not None:
            last_time, cur_time = self.__last_ckpt_time, state.elapsed_time_s
            if cur_time - last_time >= self.config.save_every_n_seconds:
                self.__last_ckpt_time = cur_time
                return True
        return False

    def save_checkpoint(self, state: State) -> Path:
        ckpt_path = self.get_ckpt_path(state)
        self.on_before_save_checkpoint(ckpt_path)

        if not is_master():
            return ckpt_path

        # Gets the path to the last checkpoint.
        logger.info("Saving checkpoint to %s", ckpt_path)
        last_ckpt_path = self.get_ckpt_path()
        ckpt_path.parent.mkdir(exist_ok=True, parents=True)

        # Potentially removes the last checkpoint.
        if last_ckpt_path.exists() and self.config.only_save_most_recent:
            if (base_ckpt := last_ckpt_path.resolve()).is_file():
                base_ckpt.unlink()

        # Saves the complete state dict to the checkpoint.
        state_dict = self.task_state_dict()
        state_dict["state"] = json.dumps(asdict(state))
        state_dict["config"] = OmegaConf.to_yaml(self.config)
        torch.save(state_dict, ckpt_path)

        # Updates the symlink to the new checkpoint.
        last_ckpt_path.unlink(missing_ok=True)
        try:
            last_ckpt_path.symlink_to(ckpt_path.relative_to(last_ckpt_path.parent))
        except FileExistsError:
            logger.exception("Exception while trying to update %s", ckpt_path)

        # Marks directory as having artifacts which shouldn't be overwritten.
        self.add_lock_file("ckpt", exists_ok=True)
        self.on_after_save_checkpoint(ckpt_path)

        return ckpt_path



"""Creates a new project, copying a standard format."""

import argparse
from pathlib import Path

import git


def get_default_author() -> str:
    try:
        return git.config.GitConfigParser().get_value("user", "name")  # type: ignore[return-value]
    except Exception:
        raise ValueError("Could not get configured Git author; please manually specify the --author argument")


def get_default_email() -> str:
    try:
        return git.config.GitConfigParser().get_value("user", "email")  # type: ignore[return-value]
    except Exception:
        raise ValueError("Could not get configured Git email; please manually specify the --email argument")


def cleanup_url(s: str) -> str:
    s = s.replace(":", "/")
    s = s.replace("git@", "https://")
    s = s.replace(".git", "")
    return s


def get_default_url() -> str:
    try:
        repo = git.Repo(search_parent_directories=True)
        return cleanup_url(repo.remote().url)
    except Exception:
        raise ValueError("Could not get configured Git URL; please manually specify the --url argument")


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("name", help="Project name")
    parser.add_argument("--description", help="Project description; if not provided, uses a generic default")
    parser.add_argument("--author", help="Project author; if not provided, defaults to the current user")
    parser.add_argument("--email", help="Project email; if not provided, defaults to <author>@users.noreply.github.com")
    parser.add_argument("--url", help="Project URL; if not provided, defaults to https://github.com/<author>/<name>")
    args = parser.parse_args()

    # Gets default values for the project metadata.
    name: str = args.name
    description: str = args.description or f"The {name} project"
    author: str = args.author or get_default_author()
    email: str = args.email or get_default_email()
    url: str = args.url or get_default_url()

    replacements = {
        "[[PROJECT NAME]]": name,
        "[[PROJECT DESCRIPTION]]": description,
        "[[PROJECT AUTHOR]]": author,
        "[[PROJECT EMAIL]]": email,
        "[[PROJECT URL]]": url,
    }

    # Copies the `assets` directory to the target directory, replacing placeholders.
    src_dir = (Path(__file__).parent / "assets").resolve()
    tgt_dir = (Path.cwd() / name).resolve()
    tgt_dir.mkdir(parents=True, exist_ok=True)

    def recursive_copy(src: Path, tgt: Path) -> None:
        if src.is_dir():
            if src.name in ("__pycache__", ".git", ".ruff_cache", ".pytest_cache"):
                return
            tgt.mkdir(parents=True, exist_ok=True)
            for src_file in src.iterdir():
                recursive_copy(src_file, tgt / src_file.name)
        else:
            with src.open("r", encoding="utf-8") as f:
                src_contents = f.read()
            tgt_contents = src_contents
            for src_str, tgt_str in replacements.items():
                tgt_contents = tgt_contents.replace(src_str, tgt_str)
            with tgt.open("w", encoding="utf-8") as f:
                f.write(tgt_contents)

    recursive_copy(src_dir, tgt_dir)

    # Renames the `project` subdirectory to the project name.
    (tgt_dir / "project").rename(tgt_dir / name)


if __name__ == "__main__":
    main()



"""Trains a simple convolutional neural network on the MNIST dataset.

Run this example with `python -m [[PROJECT NAME]].task`.
"""

from dataclasses import dataclass

import torch.nn.functional as F
from dpshdl.impl.mnist import MNIST
from torch import Tensor, nn
from torch.optim.optimizer import Optimizer

import mlfab


@dataclass
class Config(mlfab.Config):
    in_dim: int = mlfab.field(1, help="Number of input dimensions")


class MnistClassification(mlfab.Task[Config]):
    def __init__(self, config: Config) -> None:
        super().__init__(config)

        self.model = nn.Sequential(
            nn.Conv2d(config.in_dim, 32, 3, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.Conv2d(32, 32, 3, padding=1),
            nn.BatchNorm2d(32),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(32, 64, 3, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.Conv2d(64, 64, 3, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Flatten(),
            nn.Linear(64 * 7 * 7, 128),
            nn.BatchNorm1d(128),
            nn.ReLU(),
            nn.Linear(128, 10),
        )

    def get_dataset(self, phase: mlfab.Phase) -> MNIST:
        root_dir = mlfab.get_data_dir() / "mnist"
        return MNIST(root_dir=root_dir, train=phase == "train")

    def build_optimizer(self) -> Optimizer:
        return mlfab.Adam.get(self, lr=1e-3)

    def forward(self, x: Tensor) -> Tensor:
        return self.model(x)

    def get_loss(self, batch: tuple[Tensor, Tensor], state: mlfab.State) -> Tensor:
        x, y = batch
        yhat = self(x)
        self.log_step(batch, yhat, state)
        return F.cross_entropy(yhat, y.squeeze(-1).long())

    def log_valid_step(self, batch: tuple[Tensor, Tensor], output: Tensor, state: mlfab.State) -> None:
        (x, y), yhat = batch, output

        def get_label_strings() -> list[str]:
            ytrue, ypred = y.squeeze(-1), yhat.argmax(-1)
            return [f"ytrue={ytrue[i]}, ypred={ypred[i]}" for i in range(len(ytrue))]

        self.log_labeled_images("images", lambda: (x, get_label_strings()))


if __name__ == "__main__":
    MnistClassification.launch(Config(batch_size=16))



"""Defines a dummy test."""

import pytest


def test_dummy() -> None:
    assert True


@pytest.mark.slow
def test_slow() -> None:
    assert True


@pytest.mark.has_gpu
def test_gpu() -> None:
    assert True


@pytest.mark.has_mps
def test_mps() -> None:
    assert True



"""Defines general-purpose helper functions for initializing norm layers.

.. highlight:: python
.. code-block:: python

    from mlfab.nn.norms import get_norm_linear, get_norm_1d, get_norm_2d, get_norm_3d, cast_norm_type

    linear = nn.Sequential(nn.Linear(32, 32), get_norm_linear("layer", dim=32))
    conv_1d = nn.Sequential(nn.Conv1d(32, 32, 3), get_norm_1d("layer", dim=32, groups=4))
    conv_2d = nn.Sequential(nn.Conv2d(32, 32, 3), get_norm_2d("layer", dim=32, groups=4))
    conv_3d = nn.Sequential(nn.Conv3d(32, 32, 3), get_norm_3d("layer", dim=32, groups=4))

    # This lets you parametrize the norm type as a string.
    linear = nn.Sequential(nn.Linear(32, 32), get_norm_linear(cast_norm_type(my_norm), dim=32))

Choices for the norm type are:

- ``"no_norm"``: No normalization
- ``"batch"`` or ``"batch_affine"``: Batch normalization
- ``"instance"`` or ``"instance_affine"``: Instance normalization
- ``"group"`` or ``"group_affine"``: Group normalization
- ``"layer"`` or ``"layer_affine"``: Layer normalization

Note that instance norm and group norm are not available for linear layers.
"""

from typing import Literal, TypeVar, cast, get_args

import torch
from torch import Tensor, nn

T_module = TypeVar("T_module", bound=nn.Module)

NormType = Literal[
    "no_norm",
    "batch",
    "batch_affine",
    "instance",
    "instance_affine",
    "group",
    "group_affine",
    "layer",
    "layer_affine",
]

ParametrizationNormType = Literal[
    "no_norm",
    "weight",
    "spectral",
]


def cast_norm_type(s: str) -> NormType:
    args = get_args(NormType)
    assert s in args, f"Invalid norm type: '{s}' Valid options are {args}"
    return cast(NormType, s)


def cast_parametrize_norm_type(s: str) -> ParametrizationNormType:
    args = get_args(ParametrizationNormType)
    assert s in args, f"Invalid parametrization norm type: '{s}' Valid options are {args}"
    return cast(ParametrizationNormType, s)


class LastBatchNorm(nn.Module):
    """Applies batch norm along final dimension without transposing the tensor.

    The normalization is pretty simple, it basically just tracks the running
    mean and variance for each channel, then normalizes each channel to have
    a unit normal distribution.

    Input:
        x: Tensor with shape (..., N)

    Output:
        The tensor, normalized by the running mean and variance
    """

    __constants__ = ["channels", "momentum", "affine", "eps"]

    mean: Tensor
    var: Tensor

    def __init__(
        self,
        channels: int,
        momentum: float = 0.99,
        affine: bool = True,
        eps: float = 1e-4,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> None:
        super().__init__()

        self.channels = channels
        self.momentum = momentum
        self.affine = affine
        self.eps = eps

        if dtype is None:
            mean_tensor = torch.zeros(channels, device=device)
            var_tensor = torch.ones(channels, device=device)
        else:
            mean_tensor = torch.zeros(channels, device=device, dtype=dtype)
            var_tensor = torch.ones(channels, device=device, dtype=dtype)
        self.register_buffer("mean", mean_tensor)
        self.register_buffer("var", var_tensor)

        if self.affine:
            self.affine_transform = nn.Linear(channels, channels, device=device, dtype=dtype)

    def forward(self, x: Tensor) -> Tensor:
        if self.affine:
            x = self.affine_transform(x)
        if self.training:
            x_flat = x.flatten(0, -2)
            mean, var = x_flat.mean(dim=0).detach(), x_flat.var(dim=0).detach()
            new_mean = mean * (1 - self.momentum) + self.mean * self.momentum
            new_var = var * (1 - self.momentum) + self.var * self.momentum
            x_out = (x - new_mean.expand_as(x)) / (new_var.expand_as(x) + self.eps)
            self.mean.copy_(new_mean, non_blocking=True)
            self.var.copy_(new_var, non_blocking=True)
        else:
            x_out = (x - self.mean.expand_as(x)) / (self.var.expand_as(x) + self.eps)
        return x_out


class ConvLayerNorm(nn.Module):
    __constants__ = ["channels", "eps", "elementwise_affine", "static_shape"]

    def __init__(
        self,
        channels: int,
        *,
        dims: int | None = None,
        eps: float = 1e-5,
        elementwise_affine: bool = True,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> None:
        super().__init__()

        self.channels = channels
        self.eps = eps
        self.elementwise_affine = elementwise_affine

        if self.elementwise_affine:
            if dtype is None:
                self.weight = nn.Parameter(torch.empty(self.channels, device=device))
                self.bias = nn.Parameter(torch.empty(self.channels, device=device))
            else:
                self.weight = nn.Parameter(torch.empty(self.channels, device=device, dtype=dtype))
                self.bias = nn.Parameter(torch.empty(self.channels, device=device, dtype=dtype))
        else:
            self.register_parameter("weight", None)
            self.register_parameter("bias", None)

        self.static_shape = None if dims is None else (1, -1) + (1,) * dims

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if self.elementwise_affine:
            nn.init.ones_(self.weight)
            nn.init.zeros_(self.bias)

    def forward(self, inputs: Tensor) -> Tensor:
        mean = inputs.mean(dim=1, keepdim=True)
        var = torch.square(inputs - mean).mean(dim=1, keepdim=True)
        normalized_inputs = (inputs - mean) / (var + self.eps).sqrt()
        if self.elementwise_affine:
            if self.static_shape is None:
                weight = self.weight.unflatten(0, (-1,) + (1,) * (len(inputs.shape) - 2))
                bias = self.bias.unflatten(0, (-1,) + (1,) * (len(inputs.shape) - 2))
            else:
                weight = self.weight.view(self.static_shape)
                bias = self.bias.view(self.static_shape)
            normalized_inputs = normalized_inputs * weight + bias
        return normalized_inputs


def get_norm_1d(
    norm: NormType,
    *,
    dim: int | None = None,
    groups: int | None = None,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
) -> nn.Module:
    """Returns a normalization layer for tensors with shape (B, C, T).

    Args:
        norm: The norm type to use
        dim: The number of dimensions in the input tensor
        groups: The number of groups to use for group normalization
        device: The device to use for the layer
        dtype: The dtype to use for the layer

    Returns:
        A normalization layer

    Raises:
        NotImplementedError: If `norm` is not a valid 1D norm type
    """
    match norm:
        case "no_norm":
            return nn.Identity()
        case "batch" | "batch_affine":
            if dim is None:
                return nn.LazyBatchNorm1d(affine=norm == "batch_affine", device=device, dtype=dtype)
            return nn.BatchNorm1d(dim, affine=norm == "batch_affine", device=device, dtype=dtype)
        case "instance" | "instance_affine":
            if dim is None:
                return nn.LazyInstanceNorm1d(affine=norm == "instance_affine", device=device, dtype=dtype)
            return nn.InstanceNorm1d(dim, affine=norm == "instance_affine", device=device, dtype=dtype)
        case "group" | "group_affine":
            assert dim is not None, "`dim` is required for group norm"
            assert groups is not None, "`groups` is required for group norm"
            return nn.GroupNorm(groups, dim, affine=norm == "group_affine", device=device, dtype=dtype)
        case "layer" | "layer_affine":
            assert dim is not None
            return ConvLayerNorm(dim, dims=1, elementwise_affine=norm == "layer_affine", device=device, dtype=dtype)
        case _:
            raise NotImplementedError(f"Invalid 1D norm type: {norm}")


def get_norm_2d(
    norm: NormType,
    *,
    dim: int | None = None,
    groups: int | None = None,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
) -> nn.Module:
    """Returns a normalization layer for tensors with shape (B, C, H, W).

    Args:
        norm: The norm type to use
        dim: The number of dimensions in the input tensor
        groups: The number of groups to use for group normalization
        device: The device to use for the layer
        dtype: The dtype to use for the layer

    Returns:
        A normalization layer

    Raises:
        NotImplementedError: If `norm` is not a valid 2D norm type
    """
    match norm:
        case "no_norm":
            return nn.Identity()
        case "batch" | "batch_affine":
            if dim is None:
                return nn.LazyBatchNorm2d(affine=norm == "batch_affine", device=device, dtype=dtype)
            return nn.BatchNorm2d(dim, affine=norm == "batch_affine", device=device, dtype=dtype)
        case "instance" | "instance_affine":
            if dim is None:
                return nn.LazyInstanceNorm2d(affine=norm == "instance_affine", device=device, dtype=dtype)
            return nn.InstanceNorm2d(dim, affine=norm == "instance_affine", device=device, dtype=dtype)
        case "group" | "group_affine":
            assert dim is not None, "`dim` is required for group norm"
            assert groups is not None, "`groups` is required for group norm"
            return nn.GroupNorm(groups, dim, affine=norm == "group_affine", device=device, dtype=dtype)
        case "layer" | "layer_affine":
            assert dim is not None
            return ConvLayerNorm(dim, dims=2, elementwise_affine=norm == "layer_affine", device=device, dtype=dtype)
        case _:
            raise NotImplementedError(f"Invalid 2D norm type: {norm}")


def get_norm_3d(
    norm: NormType,
    *,
    dim: int | None = None,
    groups: int | None = None,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
) -> nn.Module:
    """Returns a normalization layer for tensors with shape (B, C, D, H, W).

    Args:
        norm: The norm type to use
        dim: The number of dimensions in the input tensor
        groups: The number of groups to use for group normalization
        device: The device to use for the layer
        dtype: The dtype to use for the layer

    Returns:
        A normalization layer

    Raises:
        NotImplementedError: If `norm` is not a valid 3D norm type
    """
    match norm:
        case "no_norm":
            return nn.Identity()
        case "batch" | "batch_affine":
            if dim is None:
                return nn.LazyBatchNorm3d(affine=norm == "batch_affine", device=device, dtype=dtype)
            return nn.BatchNorm3d(dim, affine=norm == "batch_affine", device=device, dtype=dtype)
        case "instance" | "instance_affine":
            if dim is None:
                return nn.LazyInstanceNorm3d(affine=norm == "instance_affine", device=device, dtype=dtype)
            return nn.InstanceNorm3d(dim, affine=norm == "instance_affine", device=device, dtype=dtype)
        case "group" | "group_affine":
            assert dim is not None, "`dim` is required for group norm"
            assert groups is not None, "`groups` is required for group norm"
            return nn.GroupNorm(groups, dim, affine=norm == "group_affine", device=device, dtype=dtype)
        case "layer" | "layer_affine":
            assert dim is not None
            return ConvLayerNorm(dim, dims=3, elementwise_affine=norm == "layer_affine", device=device, dtype=dtype)
        case _:
            raise NotImplementedError(f"Invalid 3D norm type: {norm}")


def get_norm_linear(
    norm: NormType,
    *,
    dim: int | None = None,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
) -> nn.Module:
    """Returns a normalization layer for tensors with shape (B, ..., C).

    Args:
        norm: The norm type to use
        dim: The number of dimensions in the input tensor
        device: The device to use for the layer
        dtype: The dtype to use for the layer

    Returns:
        A normalization layer

    Raises:
        NotImplementedError: If `norm` is not a valid linear norm type
    """
    match norm:
        case "no_norm":
            return nn.Identity()
        case "batch" | "batch_affine":
            assert dim is not None, "`dim` is required for batch norm"
            return LastBatchNorm(dim, affine=norm == "batch_affine", device=device, dtype=dtype)
        case "layer" | "layer_affine":
            assert dim is not None, "`dim` is required for layer norm"
            return nn.LayerNorm(dim, elementwise_affine=norm == "layer_affine", device=device, dtype=dtype)
        case _:
            raise NotImplementedError(f"Invalid linear norm type: {norm}")


def get_parametrization_norm(
    module: T_module,
    norm: ParametrizationNormType,
    *,
    name: str = "weight",
    n_power_iterations: int = 1,
    eps: float = 1e-12,
    weight_dim: int = 0,
    spectral_dim: int | None = None,
) -> T_module:
    """Returns a parametrized version of the module.

    Args:
        module: The module to parametrize
        norm: The parametrization norm type to use
        name: The name of the parameter to use for the parametrization; this
            should reference the name on the module (for instance, ``weight``
            for a ``nn.Linear`` module)
        n_power_iterations: The number of power iterations to use for spectral
            normalization
        eps: The epsilon value to use for spectral normalization
        weight_dim: The dimension of the weight parameter to normalize when
            using weight normalization
        spectral_dim: The dimension of the weight parameter to normalize when
            using spectral normalization

    Returns:
        The parametrized module
    """
    if all(p.device.type == "meta" for p in module.parameters()):
        return module
    match norm:
        case "no_norm":
            return module
        case "weight":
            return nn.utils.weight_norm(
                module,
                name=name,
                dim=weight_dim,
            )
        case "spectral":
            return nn.utils.spectral_norm(
                module,
                name=name,
                n_power_iterations=n_power_iterations,
                eps=eps,
                dim=spectral_dim,
            )
        case _:
            raise NotImplementedError(f"Invalid parametrization norm type: {norm}")



"""Defines a single interface for all the mixins."""

from mlfab.task.mixins.artifacts import ArtifactsConfig, ArtifactsMixin
from mlfab.task.mixins.checkpointing import CheckpointingConfig, CheckpointingMixin
from mlfab.task.mixins.compile import CompileConfig, CompileMixin
from mlfab.task.mixins.cpu_stats import CPUStatsConfig, CPUStatsMixin
from mlfab.task.mixins.data_loader import DataloadersConfig, DataloadersMixin
from mlfab.task.mixins.device import DeviceConfig, DeviceMixin
from mlfab.task.mixins.gpu_stats import GPUStatsConfig, GPUStatsMixin
from mlfab.task.mixins.logger import LoggerConfig, LoggerMixin
from mlfab.task.mixins.mixed_precision import MixedPrecisionConfig, MixedPrecisionMixin
from mlfab.task.mixins.optimizer import OptimizerConfig, OptimizerMixin
from mlfab.task.mixins.process import ProcessConfig, ProcessMixin
from mlfab.task.mixins.profiler import ProfilerConfig, ProfilerMixin
from mlfab.task.mixins.runnable import RunnableConfig, RunnableMixin
from mlfab.task.mixins.step_wrapper import StepContextConfig, StepContextMixin
from mlfab.task.mixins.train import TrainConfig, TrainMixin



"""Defines a distributed K-Means module.

This is used to apply K-Means clusters to a tensor. This module can be used
with cluster centers found via Scikit-Learn, Faiss, or other libraries.
"""

import logging
from typing import Callable

import numpy as np
import torch
from torch import Tensor, nn

from mlfab.nn.triton import supports_triton

logger = logging.getLogger(__name__)


def _vanilla_kmeans(x: Tensor, centers: Tensor, centers_norm: Tensor) -> Tensor:
    # Equivalent code:
    # dist = torch.norm(x[..., None, :] - centers, p=2, dim=-1)
    # return dist.argmin(dim=-1)
    x_norm = (x**2).sum(-1)
    dist = x_norm[..., None] - (2 * (x @ centers.transpose(0, 1))) + centers_norm
    # Absolute value is required here because sometimes the distance
    # can be negative due to numerical instability.
    return dist.abs().argmin(dim=-1)


def kmeans_fn(cpu: bool) -> Callable[[Tensor, Tensor, Tensor], Tensor]:
    if cpu or not supports_triton():
        return _vanilla_kmeans

    from mlfab.nn.triton.kmeans import kmeans as triton_kmeans_fn

    return triton_kmeans_fn


class KMeans(nn.Module):
    __constants__ = ["n_clusters", "n_features"]

    centers: Tensor
    centers_norm: Tensor

    def __init__(self, centers: Tensor | np.ndarray) -> None:
        super().__init__()

        n_clusters, n_features = centers.shape
        self.n_clusters = n_clusters
        self.n_features = n_features
        self.register_buffer("centers", torch.empty(n_clusters, n_features), persistent=False)
        self.register_buffer("centers_norm", torch.empty(n_clusters), persistent=False)
        self.load_centers(centers)
        self.kmeans_fn = kmeans_fn(True)
        self.kmeans_fn_cuda = kmeans_fn(False)

    def load_centers(self, centers: Tensor | np.ndarray) -> None:
        if isinstance(centers, np.ndarray):
            centers = torch.from_numpy(centers)
        assert centers.shape == self.centers.shape, f"Expected shape {self.centers.shape}, got {centers.shape}"
        self.centers.copy_(centers.to(self.centers))
        self.centers_norm.copy_((self.centers**2).sum(-1))

    def forward(self, x: Tensor) -> Tensor:
        """Applies K-Means to get cluster IDs.

        We compute ``(x - centers) ^ 2`` by rewriting as
        ``x ^ 2 - 2 * x * centers + centers ^ 2`` which avoids expanding the
        tensor when doing the norm.

        Args:
            x: The input tensor, with shape ``(*, n_features)``

        Returns:
            The cluster IDs, with shape ``(*)``
        """
        kmeans_fn = self.kmeans_fn_cuda if x.is_cuda else self.kmeans_fn
        return kmeans_fn(x, self.centers, self.centers_norm)



# mypy: disable-error-code="import"
"""Defines utility function for Triton."""

import functools

import torch

from mlfab.core.conf import load_user_config
from mlfab.nn.device.gpu import gpu_device
from mlfab.utils.text import show_warning


@functools.lru_cache(maxsize=None)
def supports_triton() -> bool:
    config = load_user_config().triton
    if not config.use_triton_if_available:
        return False

    if not gpu_device.has_device():
        return False

    try:
        import triton

        assert triton is not None
        return True
    except (ImportError, ModuleNotFoundError):
        if gpu_device.has_device():
            show_warning("Triton is not installed, but CUDA is available; install with `pip install triton`")
        return False



# mypy: disable-error-code="override"
"""Defines the one-to-many monotonic attention forward and backward passes in log-space PyTorch."""

import functools
import math
import warnings
from typing import Callable, Literal, TypeVar

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.autograd.function import Function, FunctionCtx, once_differentiable

from mlfab.nn.triton import supports_triton

MIN_LOG_PROB = -1e4


def _logaddexp(*ts: Tensor) -> Tensor:
    return torch.logsumexp(torch.stack(ts, dim=-1), dim=-1)


def _pos_log_prob(x: Tensor) -> Tensor:
    return -torch.log1p(torch.exp(-x))


def _neg_log_prob(x: Tensor) -> Tensor:
    return -torch.log1p(torch.exp(x))


def _d_pos_log_prob(x: Tensor) -> Tensor:
    return 1 / (1 + torch.exp(x))


def _d_neg_log_prob(x: Tensor) -> Tensor:
    return -1 / (1 + torch.exp(-x))


def forward_pass_cpu_(logits: Tensor) -> tuple[Tensor, Tensor]:
    t_i = logits.size(-2)
    logits = F.pad(logits, (1, 0), value=0.0)
    phis = torch.empty_like(logits)
    phis[..., :, 0] = MIN_LOG_PROB
    phis[..., 0, :] = MIN_LOG_PROB
    phis[..., 0, 1] = 0.0
    for i in range(1, t_i):
        phis[..., i, 1:] = _logaddexp(
            phis[..., i - 1, 1:] + _pos_log_prob(logits[..., i - 1, 1:]),
            phis[..., i - 1, :-1] + _neg_log_prob(logits[..., i - 1, :-1]),
        )
    phis = phis[..., 1:]
    return phis, phis + _neg_log_prob(logits[..., 1:])


def backward_pass_cpu_(logits: Tensor, phis: Tensor, grad_phis: Tensor) -> Tensor:
    t_i = logits.size(-2)
    grad_logits = torch.empty_like(grad_phis)
    grad_logits[..., t_i - 1, :] = 0.0
    extra_grad = _d_neg_log_prob(logits) * grad_phis
    for i in range(t_i - 2, -1, -1):
        p = phis[..., i + 1, :]
        a = (phis[..., i, :] + _pos_log_prob(logits[..., i, :]) - p).exp()
        b = (phis[..., i, :-1] + _neg_log_prob(logits[..., i, :-1]) - p[..., 1:]).exp()
        c = grad_phis[..., i + 1, :] * a
        d = grad_phis[..., i + 1, 1:] * b
        grad_logits[..., i, :] = c * _d_pos_log_prob(logits[..., i, :])
        grad_logits[..., i, :-1] += d * _d_neg_log_prob(logits[..., i, :-1])
        grad_phis[..., i, :] += c
        grad_phis[..., i, :-1] += d
    return grad_logits + extra_grad


class MonotonicAttentionCpu(Function):
    @staticmethod
    def forward(ctx: FunctionCtx, logits: Tensor) -> Tensor:
        phis, probs = forward_pass_cpu_(logits)
        ctx.save_for_backward(logits, phis)
        return probs

    @staticmethod
    @once_differentiable
    def backward(ctx: FunctionCtx, grad_phis: Tensor) -> Tensor:
        logits, phis = ctx.saved_tensors
        grad_logits = backward_pass_cpu_(logits, phis, grad_phis.clone())
        return grad_logits


def monotonic_attention_cpu(logits: Tensor) -> Tensor:
    """Computes the monotonic attention normalization on the transition probabilities.

    Args:
        logits: The transition logits, with shape ``(bsz, tsz_src, tsz_tgt)``

    Returns:
        The marginalized log probabilities for each cell being part of a
        monotonic alignment path, with shape ``(bsz, tsz_src, tsz_tgt)``.
    """
    tsz_src, tsz_tgt = logits.size(-2), logits.size(-1)
    if tsz_tgt > tsz_src:
        warnings.warn("One-to-many attention expects the source sequence to be longer than the target sequence!")
    return MonotonicAttentionCpu.apply(logits)


Mode = Literal["many_keys_one_query", "one_key_many_queries"]


Tq = TypeVar("Tq", Tensor, None)
Tk = TypeVar("Tk", Tensor, None)
Tv = TypeVar("Tv", Tensor, None)


def phi_to_pos_prob(phi: float) -> float:
    return 1 / (math.exp(-phi) + 1)


def phi_to_neg_prob(phi: float) -> float:
    return 1 / (math.exp(phi) + 1)


def pos_prob_to_phi(prob: float) -> float:
    return math.log((1 / prob) - 1)


def neg_prob_to_phi(prob: float) -> float:
    return -math.log((1 / prob) - 1)


@functools.lru_cache(maxsize=None)
def get_monotonic_attention_fn(device_type: str) -> Callable[[Tensor], Tensor]:
    if device_type != "cuda" or not supports_triton():
        return monotonic_attention_cpu

    from mlfab.nn.triton.monotonic_attention import monotonic_attention_gpu

    return monotonic_attention_gpu


def monotonic_attention(logits: Tensor) -> Tensor:
    """Computes the monotonic attention normalization on the transition probabilities.

    Args:
        logits: The transition logits, with shape ``(bsz, tsz_src, tsz_tgt)``

    Returns:
        The marginalized log probabilities for each cell being part of a
        monotonic alignment path, with shape ``(bsz, tsz_src, tsz_tgt)``.
    """
    return get_monotonic_attention_fn(logits.device.type)(logits)


class MonotonicAttention(nn.Module):
    """Defines a one-to-many multihead monotonic attention layer.

    Parameters:
        mode: Specifies either many keys for each query, or many queries for
            each key. In the former case, the key and value sequences should
            be longer than the query sequence; vice versa in the latter case.
        embed_dim: The input and output embedding dimension.
        num_heads: The number of attention heads.
        bias: Whether to include a bias term in the projection layers.
        kdim: The dimension of the key projection. Defaults to ``embed_dim``.
        vdim: The dimension of the value projection. Defaults to ``embed_dim``.
        gqa_factor: The GQA factor to use, meaning the ratio of the number of
            queries to the number of keys. Higher values will result in more
            queries than keys, which can speed up inference.
        clamp_prob: The maximum transition probability to clamp to.
        soft_clamp: If set, soft-clamp the transition probabailities using the
            tanh function, otherwise hard-clamp them.

    Inputs:
        query: The query tensor, of shape ``(B, T, C)``.
        key: The key tensor, of shape ``(B, T, C)``.
        value: The value tensor, of shape ``(B, T, C)``.
        state: The previous key and value tensors, of shape
            ``(B * H, T', C // H)``, where ``T'`` is the number of previous
            timesteps and ``H`` is the number of attention heads. This is
            only supported if ``is_causal=True``.
        is_causal: Whether to apply a causal mask to the attention matrix.
            Note that the "mask" is only applied implicitly and isn't actually
            instantiated as a tensor.

    Outputs:
        output: The output tensor, of shape ``(B, T, C)``, along with the
            key and value state for the next timestep.
    """

    __constants__ = [
        "num_heads",
        "gqa_factor",
        "kv_num_heads",
        "head_dim",
        "max_kv_cache_len",
        "embed_dim",
        "kv_embed_dim",
        "kdim",
        "vdim",
        "_qkv_same_embed_dim",
        "clamp_prob",
        "soft_clamp",
        "norm_fact",
    ]

    def __init__(
        self,
        mode: Mode,
        embed_dim: int,
        num_heads: int = 1,
        bias: bool = True,
        kdim: int | None = None,
        vdim: int | None = None,
        gqa_factor: int = 1,
        max_kv_cache_len: int | None = None,
        clamp_prob: float = 0.98,
        soft_clamp: bool = True,
    ) -> None:
        super().__init__()

        assert embed_dim % num_heads == 0, f"`{embed_dim=}` must be divisible by `{num_heads=}`"
        assert num_heads % gqa_factor == 0, f"`{num_heads=}` must be divisible by `{gqa_factor=}`"
        assert clamp_prob > 0.5 and clamp_prob < 1.0, f"`{clamp_prob=}` must be in the range (0.5, 1.0]"

        # Stores some constant values.
        self.mode = mode
        self.num_heads = num_heads
        self.gqa_factor = gqa_factor
        self.kv_num_heads = num_heads // gqa_factor
        self.head_dim = embed_dim // num_heads
        self.max_kv_cache_len = max_kv_cache_len
        self.clamp_value = neg_prob_to_phi(clamp_prob)
        self.soft_clamp = soft_clamp
        self.norm_fact = math.sqrt(self.head_dim)

        self.embed_dim = embed_dim
        self.kv_embed_dim = self.kv_num_heads * self.head_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self._qkv_same_embed_dim = self.kdim == embed_dim and self.vdim == embed_dim

        if not self._qkv_same_embed_dim:
            self.q_proj_weight = nn.Parameter(torch.empty((embed_dim, embed_dim)))
            self.k_proj_weight = nn.Parameter(torch.empty((self.kv_embed_dim, self.kdim)))
            self.v_proj_weight = nn.Parameter(torch.empty((self.kv_embed_dim, self.vdim)))
            self.register_parameter("in_proj_weight", None)
        else:
            self.in_proj_weight = nn.Parameter(torch.empty((embed_dim + self.kv_embed_dim * 2, embed_dim)))
            self.register_parameter("q_proj_weight", None)
            self.register_parameter("k_proj_weight", None)
            self.register_parameter("v_proj_weight", None)

        if bias:
            self.in_proj_bias = nn.Parameter(torch.empty(embed_dim + self.kv_embed_dim * 2))
        else:
            self.register_parameter("in_proj_bias", None)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)

        self._reset_parameters()

    def _reset_parameters(self) -> None:
        if self._qkv_same_embed_dim:
            nn.init.xavier_uniform_(self.in_proj_weight)
        else:
            nn.init.xavier_uniform_(self.q_proj_weight)
            nn.init.xavier_uniform_(self.k_proj_weight)
            nn.init.xavier_uniform_(self.v_proj_weight)

        if self.in_proj_bias is not None:
            nn.init.constant_(self.in_proj_bias, 0.0)
            nn.init.constant_(self.out_proj.bias, 0.0)

    def _clamp_attn(self, attn: Tensor) -> Tensor:
        if not self.soft_clamp:
            return attn.clamp(-self.clamp_value, self.clamp_value)
        return attn.tanh() * self.clamp_value

    def _compute_attn(self, xq: Tensor, xk: Tensor, mask: Tensor | None = None) -> Tensor:
        bsz, gqa, num_heads = xq.shape[:3]

        if self.mode == "one_key_many_queries":
            attn = torch.einsum("bghqc,bghkc->bghqk", xq, xk) / self.norm_fact
            if mask is not None:
                attn = attn + mask[:, None, None]
            attn = self._clamp_attn(attn)
            return monotonic_attention(attn.flatten(0, 2)).unflatten(0, (bsz, gqa, num_heads))

        if self.mode == "many_keys_one_query":
            attn = torch.einsum("bghqc,bghkc->bghkq", xq, xk) / self.norm_fact
            if mask is not None:
                attn = attn + mask.transpose(-2, -1)[:, None, None]
            attn = self._clamp_attn(attn)
            output = monotonic_attention(attn.flatten(0, 2)).unflatten(0, (bsz, gqa, num_heads))
            return output.transpose(-2, -1)

        raise NotImplementedError(f"Unknown mode: {self.mode}")

    def forward_matmuls(self, query: Tq, key: Tk, value: Tv) -> tuple[Tq, Tk, Tv]:
        # Gets the weights for the matmuls.
        qkw_splits = (self.embed_dim, self.kv_embed_dim, self.kv_embed_dim)
        if self._qkv_same_embed_dim:
            qw, kw, vw = self.in_proj_weight.split(qkw_splits, dim=0)
        else:
            qw, kw, vw = self.q_proj_weight, self.k_proj_weight, self.v_proj_weight
        if self.in_proj_bias is None:
            qb, kb, vb = None, None, None
        else:
            qb, kb, vb = self.in_proj_bias.split(qkw_splits, dim=0)

        # Computes the query projection.
        if query is None:
            xq = None
        else:
            assert query.dim() == 3
            xq = F.linear(query, qw, qb)
            xq = xq.unflatten(-1, (self.gqa_factor, self.kv_num_heads, self.head_dim)).permute(0, 2, 3, 1, 4)

        # Computes the key projection.
        if key is None:
            xk = None
        else:
            assert key.dim() == 3
            xk = F.linear(key, kw, kb)
            xk = xk.unflatten(-1, (1, self.kv_num_heads, self.head_dim)).permute(0, 2, 3, 1, 4)

        # Computes the value projection.
        if value is None:
            xv = None
        else:
            assert value.dim() == 3
            xv = F.linear(value, vw, vb)
            xv = xv.unflatten(-1, (1, self.kv_num_heads, self.head_dim)).permute(0, 2, 3, 1, 4)

        return xq, xk, xv

    def get_attn_matrix(self, query: Tensor, key: Tensor, mask: Tensor | None = None) -> Tensor:
        """Computes the attention matrix for a given query and key.

        This function can be used for visualization purposes.

        Args:
            query: The query vector, with shape ``(B, Tq, C)``
            key: The key vector, with shape ``(B, Tk, C)``
            mask: The attention mask, of shape ``(B, Tq, Tk)``. If ``None``,
                don't apply an attention mask.

        Returns:
            The attention matrix, of shape ``(B, G, H, Tq, Tk)``.
        """
        xq, xk, _ = self.forward_matmuls(query, key, None)
        return self._compute_attn(xq, xk, mask)

    def forward_attn(self, xq: Tensor, xk: Tensor, xv: Tensor, mask: Tensor | None = None) -> Tensor:
        attn = self._compute_attn(xq, xk, mask).softmax(dim=-1)
        xo = torch.einsum("bghqk,bghkc->bghqc", attn, xv)
        xo = xo.permute(0, 3, 1, 2, 4).flatten(2)
        xo = self.out_proj(xo)
        return xo

    def forward(self, query: Tensor, key: Tensor, value: Tensor, mask: Tensor | None = None) -> Tensor:
        xq, xk, xv = self.forward_matmuls(query, key, value)
        xo = self.forward_attn(xq, xk, xv, mask)
        return xo



"""Defines a bunch of dataset transforms."""

import random
from typing import TypeVar

import torch
import torchvision.transforms.functional as V
from PIL.Image import Image as PILImage
from torch import Tensor, nn
from torchvision.transforms import InterpolationMode

Image = TypeVar("Image", Tensor, PILImage)

NormParams = tuple[float, float, float]

# Default image normalization parameters.
MEAN: NormParams = 0.48145466, 0.4578275, 0.40821073
STD: NormParams = 0.26862954, 0.26130258, 0.27577711


def square_crop(img: Image) -> Image:
    """Crops an image to a square.

    Args:
        img: The input image

    Returns:
        The cropped image, with height and width equal.
    """
    img_width, img_height = V.get_image_size(img)
    height = width = min(img_height, img_width)
    top, left = (img_height - height) // 2, (img_width - width) // 2
    return V.crop(img, top, left, height, width)


def square_resize_crop(img: Image, size: int, interpolation: InterpolationMode = InterpolationMode.NEAREST) -> Image:
    """Resizes an image to a square and then crops it.

    Args:
        img: The input image
        size: The size of the square
        interpolation: The interpolation mode to use

    Returns:
        The cropped image
    """
    img_width, img_height = V.get_image_size(img)
    min_dim = min(img_width, img_height)
    height, width = int((img_width / min_dim) * size), int((img_height / min_dim) * size)
    img = V.resize(img, [height, width], interpolation)
    top, left = (height - size) // 2, (width - size) // 2
    return V.crop(img, top, left, size, size)


def upper_left_crop(img: Image, height: int, width: int) -> Image:
    """Crops an image from the upper left corner.

    This is useful because it preserves camera intrinsics for an image.

    Args:
        img: The input image
        height: The height of the crop
        width: The width of the crop

    Returns:
        The cropped image
    """
    return V.crop(img, 0, 0, height, width)


def normalize(t: Tensor, *, mean: NormParams = MEAN, std: NormParams = STD) -> Tensor:
    """Normalizes an image tensor (by default, using ImageNet parameters).

    This can be paired with :func:`denormalize` to convert an image tensor
    to a normalized tensor for processing by a model.

    Args:
        t: The input tensor
        mean: The mean to subtract
        std: The standard deviation to divide by

    Returns:
        The normalized tensor
    """
    return V.normalize(t, mean, std)


def denormalize(t: Tensor, *, mean: NormParams = MEAN, std: NormParams = STD) -> Tensor:
    """Denormalizes a tensor.

    This can be paired with :func:`normalize` to convert a normalized tensor
    back to the original image for viewing by humans.

    Args:
        t: The input tensor
        mean: The mean to subtract
        std: The standard deviation to divide by

    Returns:
        The denormalized tensor
    """
    mean_tensor = torch.tensor(mean, device=t.device, dtype=t.dtype)
    std_tensor = torch.tensor(std, device=t.device, dtype=t.dtype)
    return (t * std_tensor[None, :, None, None]) + mean_tensor[None, :, None, None]


def random_square_crop(img: Image) -> Image:
    """Randomly crops an image to a square.

    Args:
        img: The input image

    Returns:
        The cropped image
    """
    img_width, img_height = V.get_image_size(img)
    height = width = min(img_height, img_width)
    top, left = random.randint(0, img_height - height), random.randint(0, img_width - width)
    return V.crop(img, top, left, height, width)


def random_square_crop_multi(imgs: list[Image]) -> list[Image]:
    """Randomly crops a list of images to the same size.

    Args:
        imgs: The list of images to crop

    Returns:
        The cropped images
    """
    img_dims = V.get_image_size(imgs[0])
    assert all(V.get_image_size(i) == img_dims for i in imgs[1:])
    img_width, img_height = img_dims
    height = width = min(img_width, img_height)
    top, left = random.randint(0, img_height - height), random.randint(0, img_width - width)
    return [V.crop(i, top, left, height, width) for i in imgs]


def make_size(img: Image, ref_size: tuple[int, int]) -> Image:
    """Converts an image to a specific size, zero-padding smaller dimension.

    Args:
        img: The input image
        ref_size: The reference size, as (width, height)

    Returns:
        The resized image
    """
    img_c, (img_w, img_h), (ref_w, ref_h) = V.get_image_num_channels(img), V.get_image_size(img), ref_size
    if img_h / img_w < ref_h / ref_w:  # Pad width
        new_h, new_w = (img_h * ref_w) // img_w, ref_w
    else:
        new_h, new_w = ref_h, (img_w * ref_h) // img_h
    img = V.resize(img, [new_h, new_w], InterpolationMode.BILINEAR)
    new_img = img.new_zeros(img_c, ref_h, ref_w)
    start_h, start_w = (ref_h - new_h) // 2, (ref_w - new_w) // 2
    new_img[:, start_h : start_h + new_h, start_w : start_w + new_w] = img
    return new_img


def make_same_size(img: Image, ref_img: Image) -> Image:
    """Converts an image to the same size as a reference image.

    Args:
        img: The input image
        ref_img: The reference image

    Returns:
        The input image resized to the same size as the reference image,
            zero-padding dimensions which are too small
    """
    ref_w, ref_h = V.get_image_size(ref_img)
    return make_size(img, (ref_w, ref_h))


def pil_to_tensor(pic: PILImage) -> Tensor:
    """Converts a PIL image to a tensor.

    Args:
        pic: The PIL image.

    Returns:
        The normalized image tensor.
    """
    tensor = V.pil_to_tensor(pic)
    tensor = V.convert_image_dtype(tensor)
    if tensor.shape[0] == 3:
        tensor = normalize(tensor)
    return tensor


class SquareResizeCrop(nn.Module):
    """Resizes and crops an image to a square with the target shape.

    Generally SquareCrop followed by a resize should be preferred when using
    bilinear resize, as it is faster to do the interpolation on the smaller
    image. However, nearest neighbor resize on the larger image followed by
    a crop on the smaller image can sometimes be faster.
    """

    __constants__ = ["size", "interpolation"]

    def __init__(self, size: int, interpolation: InterpolationMode = InterpolationMode.NEAREST) -> None:
        """Initializes the square resize crop.

        Args:
            size: The square height and width to resize to
            interpolation: The interpolation type to use when resizing
        """
        super().__init__()

        self.size = int(size)
        self.interpolation = InterpolationMode(interpolation)

    def forward(self, img: Image) -> Image:
        return square_resize_crop(img, self.size, self.interpolation)


class UpperLeftCrop(nn.Module):
    """Crops image from upper left corner, to preserve image intrinsics."""

    __constants__ = ["height", "width"]

    def __init__(self, height: int, width: int) -> None:
        """Initializes the upper left crop.

        Args:
            height: The max height of the cropped image
            width: The max width of the cropped image
        """
        super().__init__()

        self.height, self.width = height, width

    def forward(self, img: Image) -> Image:
        return upper_left_crop(img, self.height, self.width)



"""Defines PyTest configuration for the project."""

import functools
import random

import numpy as np
import pytest
import torch
from _pytest.python import Function, Metafunc


@pytest.fixture(autouse=True)
def set_random_seed() -> None:
    random.seed(1337)
    np.random.seed(1337)
    torch.manual_seed(1337)


@functools.lru_cache()
def has_gpu() -> bool:
    return torch.cuda.is_available()


@functools.lru_cache()
def has_mps() -> bool:
    return torch.backends.mps.is_available()


def pytest_runtest_setup(item: Function) -> None:
    for mark in item.iter_markers():
        if mark.name == "has_gpu" and not has_gpu():
            pytest.skip("Skipping because this test requires a GPU and none is available")


def pytest_collection_modifyitems(items: list[Function]) -> None:
    items.sort(key=lambda x: x.get_closest_marker("slow") is not None)


def pytest_generate_tests(metafunc: Metafunc) -> None:
    if "device" in metafunc.fixturenames:
        torch_devices = [torch.device("cpu")]
        if has_gpu():
            torch_devices.append(torch.device("cuda"))
        if has_mps():
            torch_devices.append(torch.device("mps"))
        metafunc.parametrize("device", torch_devices)



"""Defines the base launcher class."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from mlfab.task.base import RawConfigType

if TYPE_CHECKING:
    from mlfab.task.mixins.runnable import Config, RunnableMixin


class BaseLauncher(ABC):
    """Defines the base launcher class."""

    @abstractmethod
    def launch(
        self,
        task: "type[RunnableMixin[Config]]",
        *cfgs: RawConfigType,
        use_cli: bool | list[str] = True,
    ) -> None:
        """Launches the training process.

        Args:
            task: The task class to train
            cfgs: The raw configuration to use for training
            use_cli: Whether to include CLI arguments in the configuration
        """



"""Defines a launcher to launch a Slurm training job."""

import argparse
import datetime
import functools
import json
import os
import re
import signal
import subprocess
import sys
import warnings
from dataclasses import dataclass
from pathlib import Path

import torch

from mlfab.nn.parallel import (
    init_parallelism,
    init_process_group_from_backend,
    is_master,
    set_init_method,
    set_master_addr,
    set_master_port,
    set_rank,
    set_world_size,
)
from mlfab.task.base import RawConfigType
from mlfab.task.launchers.staged import StagedLauncher
from mlfab.task.mixins.artifacts import ArtifactsMixin, Config as ArtifactsConfig
from mlfab.task.mixins.runnable import Config as RunnableConfig, RunnableMixin
from mlfab.utils.logging import configure_logging
from mlfab.utils.text import show_info

DEFAULT_MASTER_PORT = 29500


def set_slurm_rank_and_world_size() -> tuple[int, int]:
    node_id = int(os.environ["SLURM_NODEID"])
    local_id = int(os.environ["SLURM_LOCALID"])
    tasks_per_node = int(os.environ["SLURM_NTASKS_PER_NODE"])
    num_nodes = int(os.environ["SLURM_NNODES"])
    rank = node_id * tasks_per_node + local_id
    world_size = num_nodes * tasks_per_node
    set_rank(rank)
    set_world_size(world_size)
    return rank, world_size


def set_slurm_master_addr_and_port() -> str:
    node_list = os.environ.get("SLURM_STEP_NODELIST")
    if node_list is None:
        node_list = os.environ.get("SLURM_JOB_NODELIST")
    assert node_list is not None, "`SLURM_JOB_NODELIST` environment variable not set"
    hostnames = subprocess.check_output(["scontrol", "show", "hostnames", node_list])
    host = hostnames.split()[0].decode("utf-8")
    port = int(os.environ.get("MASTER_PORT", str(DEFAULT_MASTER_PORT)))
    set_master_addr(host)
    set_master_port(port)
    return host


def write_message(message: str) -> None:
    sys.stderr.write(message)
    sys.stderr.flush()


def requeue_job() -> None:
    if is_master():
        if "SLURM_JOB_ID" in os.environ:
            cmd = ["scontrol", "requeue", os.environ["SLURM_JOB_ID"]]
            write_message(f"Requeuing job {os.environ['SLURM_JOB_ID']}\n")
            subprocess.check_call(cmd)
        else:
            write_message("SLURM_JOB_ID environment variable not found; not requeueing\n")


@dataclass
class PartitionInfo:
    name: str
    gpus_per_node: int
    cpus_per_node: int
    time_limit: str


@functools.lru_cache()
def parse_sinfo_output() -> list[PartitionInfo]:
    sinfo_output = subprocess.check_output(["sinfo", "--format", "%c %G %P %l"])
    partition_infos: list[PartitionInfo] = []
    lines = sinfo_output.decode("utf-8").splitlines()[1:]
    for line in lines:
        cpus_per_node, gres, name, time_limit = line.split()

        # Parses GPUs per node from gres.
        gpus_per_node_re = re.search(r"gpu:(\d+)", gres)
        if gpus_per_node_re is None:
            continue
        gpus_per_node = int(gpus_per_node_re.group(1))

        # Cleans up partition name.
        name = name.replace("*", "")

        partition_infos += [PartitionInfo(name, int(gpus_per_node), int(cpus_per_node), time_limit)]

    return partition_infos


@dataclass
class SlurmArgs:
    partition: str | None
    gpus_per_node: int | None
    num_nodes: int
    num_jobs: int
    account: str | None


class SlurmLauncher(StagedLauncher):
    """Defines a launcher to launch a Slurm training job.

    If no parameters are supplied, this launcher will attempt to use `sinfo` to
    find a partition with at least one GPU, and will use the first such
    partition found, with the default number of GPUs per node and CPUs per GPU.
    """

    def __init__(
        self,
        partition: str | None = None,
        gpus_per_node: int | None = None,
        cpus_per_gpu: int | None = None,
        num_nodes: int = 1,
        gpu_type: str | None = None,
        exclusive: bool = False,
        time_limit: str | None = None,
        num_jobs: int = 1,
        comment: str | None = None,
        master_port: int = DEFAULT_MASTER_PORT,
        model_parallelism: int = 1,
        pipeline_parallelism: int = 1,
        backend: str | None = None,
        model_parallel_backend: str | None = None,
        pipeline_parallel_backend: str | None = None,
        data_parallel_backend: str | None = None,
        account: str | None = None,
    ) -> None:
        super().__init__()

        if partition is None:
            if len(sinfo_output := parse_sinfo_output()) == 0:
                raise RuntimeError("`sinfo` did not return any partitions with available GPUs!")
            partition = sinfo_output[0].name

        if gpus_per_node is None or cpus_per_gpu is None:
            try:
                first_partition = next(p for p in parse_sinfo_output() if p.name == partition)
            except StopIteration:
                raise RuntimeError(f"Partition {partition} not found in `sinfo` output")
            if gpus_per_node is None:
                gpus_per_node = first_partition.gpus_per_node
            if cpus_per_gpu is None:
                cpus_per_gpu = first_partition.cpus_per_node // gpus_per_node

        self.partition: str = partition
        self.gpus_per_node: int = gpus_per_node
        self.cpus_per_gpu: int = cpus_per_gpu

        self.num_nodes = num_nodes
        self.gpu_type = gpu_type
        self.exclusive = exclusive
        self.time_limit = time_limit
        self.num_jobs = num_jobs
        self.comment = comment
        self.master_port = master_port
        self.model_parallelism = model_parallelism
        self.pipeline_parallelism = pipeline_parallelism
        self.backend = backend
        self.model_parallel_backend = model_parallel_backend
        self.pipeline_parallel_backend = pipeline_parallel_backend
        self.data_parallel_backend = data_parallel_backend
        self.account = account

    @classmethod
    def parse_args_from_cli(cls) -> tuple[SlurmArgs, list[str]]:
        parser = argparse.ArgumentParser(description="Launches a Slurm job.")
        parser.add_argument("--partition", type=str, default=None, help="The partition to use")
        parser.add_argument("--gpus-per-node", type=int, default=None, help="The number of GPUs per node")
        parser.add_argument("--num-nodes", type=int, default=1, help="The number of nodes to use")
        parser.add_argument("--num-jobs", type=int, default=1, help="The number of jobs to launch")
        parser.add_argument("--account", type=str, default=None, help="The account to use")
        args, remaining_args = parser.parse_known_intermixed_args()

        return (
            SlurmArgs(
                partition=args.partition,
                gpus_per_node=args.gpus_per_node,
                num_nodes=args.num_nodes,
                num_jobs=args.num_jobs,
                account=args.account,
            ),
            remaining_args,
        )

    @property
    def extra_sbatch_lines(self) -> list[str]:
        sbatch_lines: list[str] = []
        if "EMAIL" in os.environ:
            sbatch_lines += [f"--mail-user={os.environ['EMAIL']}", "--mail-type=ALL"]
        if self.account is not None:
            sbatch_lines += [f"--account={self.account}"]
        if self.exclusive:
            sbatch_lines += ["--exclusive"]
        if self.time_limit is not None:
            sbatch_lines += [f"--time={self.time_limit}"]
        return sbatch_lines

    @property
    def extra_export_lines(self) -> str:
        export_lines: dict[str, str] = {}
        if self.model_parallelism > 1:
            export_lines["MODEL_PARALLELISM"] = str(self.model_parallelism)
        if self.pipeline_parallelism > 1:
            export_lines["PIPELINE_PARALLELISM"] = str(self.pipeline_parallelism)
        if self.backend is not None:
            export_lines["BACKEND"] = self.backend
        if self.model_parallel_backend is not None:
            export_lines["MODEL_PARALLEL_BACKEND"] = self.model_parallel_backend
        if self.pipeline_parallel_backend is not None:
            export_lines["PIPELINE_PARALLEL_BACKEND"] = self.pipeline_parallel_backend
        if self.data_parallel_backend is not None:
            export_lines["DATA_PARALLEL_BACKEND"] = self.data_parallel_backend
        return "\n".join(f"export {k}={v}" for k, v in sorted(export_lines.items()))

    def pythonpath(self, stage_dir: str | Path | None) -> str:
        pythonpath_paths = ([] if stage_dir is None else [str(stage_dir)]) + os.environ.get("PYTHONPATH", "").split(":")
        return ":".join(p for p in pythonpath_paths if p)

    def sbatch_file_contents(self, task: "ArtifactsMixin[ArtifactsConfig]") -> str:
        output_path, error_path = task.exp_dir / "slurm.out", task.exp_dir / "slurm.err"
        stage_dir = task.stage_environment()
        comments = ([] if self.comment is None else [self.comment]) + [f"Log directory: {task.exp_dir}"]
        config_path = self.get_config_path(task, use_cli=False)

        # Gets the extra sbatch lines.
        extra_sbatch_lines = self.extra_sbatch_lines
        if stage_dir is not None:
            extra_sbatch_lines += [f"--chdir={stage_dir}"]
            comments += [f"Code location: {stage_dir}"]
        extra_sbatch_lines_str = "\n".join(f"#SBATCH {line}" for line in extra_sbatch_lines)

        return f"""
#!/bin/bash
#SBATCH --job-name={task.task_name}
#SBATCH --partition={self.partition}
#SBATCH --requeue
#SBATCH --signal=USR1@60
#SBATCH --comment='{'; '.join(comments)}'
#SBATCH --nodes={self.num_nodes}
#SBATCH --ntasks-per-node={self.gpus_per_node}
#SBATCH --cpus-per-gpu={self.cpus_per_gpu}
#SBATCH --gpus-per-node={self.gpus_per_node}
#SBATCH --output={output_path}
#SBATCH --error={error_path}
#SBATCH --open-mode=append
{extra_sbatch_lines_str}

# Sets the environment variables.
export SLURM_EXPORT_ENV=ALL
export PYTHONPATH={self.pythonpath(stage_dir)}
export MASTER_PORT={self.master_port}
{self.extra_export_lines}

# Set some debugging flags.
export TORCH_DISTRIBUTED_DEBUG=DETAIL
export TORCH_SHOW_CPP_STACKTRACES=1
export NCCL_DEBUG=1

# Make a new line in the stdout file.
echo ""
echo "***"
echo "Job ID: ${{SLURM_JOBID}} - $(date)"
echo "***"
echo ""

# Also make a new line in the stderr file.
echo "" >&2
echo "***" >&2
echo "Job ID: ${{SLURM_JOBID}} - $(date)" >&2
echo "***" >&2
echo "" >&2

# Runs the training command.
srun \\
    --nodes={self.num_nodes} \\
    --ntasks-per-node={self.gpus_per_node} \\
    --cpus-per-gpu={self.cpus_per_gpu} \\
    --gpus-per-node={self.gpus_per_node} \\
    python -m {self.__module__} {task.task_key} {config_path}
""".strip()

    def update_job_info(self, task: ArtifactsMixin, all_run_ids: list[str]) -> None:
        job_file = task.exp_dir / "slurm_info.json"

        # Loads existing job information, or creates an empty list.
        job_info: list
        if job_file.exists():
            with open(job_file, "r", encoding="utf-8") as f:
                job_info = json.load(f)
        else:
            job_info = []

        # Adds the new job information.
        job_info += [
            {
                "launch_time": datetime.datetime.now().isoformat(),
                "job_ids": all_run_ids,
                "task_key": task.task_key,
                "exp_dir": str(task.exp_dir),
            },
        ]

        # Writes the updated job information to a file.
        with open(job_file, "w", encoding="utf-8") as f:
            json.dump(job_info, f, indent=2)

    def launch(
        self,
        task: "type[RunnableMixin[RunnableConfig]]",
        *cfgs: RawConfigType,
        use_cli: bool | list[str] = True,
    ) -> None:
        if not issubclass(task, ArtifactsMixin):
            raise RuntimeError(f"Task {task} must be an `ArtifactsMixin`")

        # Creates the task using the meta device to avoid instantiating weights.
        with torch.device("meta"), warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            task_obj = task.get_task(*cfgs, use_cli=use_cli)

        # Writes the sbatch file.
        sbatch_path = task_obj.exp_dir / "sbatch.sh"
        with open(sbatch_path, "w", encoding="utf-8") as f:
            f.write(self.sbatch_file_contents(task_obj))

        # Calls `sbatch` on the given file.
        all_run_ids: list[str] = []
        for _ in range(self.num_jobs):
            command = ["sbatch", str(sbatch_path)]
            if all_run_ids:
                command += ["--dependency", all_run_ids[-1]]
            proc = subprocess.Popen(  # pylint: disable=consider-using-with
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            assert proc is not None and proc.stdout is not None
            proc.wait()
            log_line = proc.stdout.read().decode("utf-8").strip()
            run_ids = re.findall(r"Submitted batch job (\d+)", log_line)
            assert len(run_ids) == 1, f"Unexpected log line: {log_line}"
            all_run_ids += [run_ids[0]]

        run_ids_str = "".join(f"\n - {run_id}" for run_id in all_run_ids)
        show_info(f"Launched {len(all_run_ids)} job(s) to {task_obj.exp_dir}:{run_ids_str}")

        # Writes the job information to a file.
        self.update_job_info(task_obj, all_run_ids)

        task_obj.add_lock_file("scheduled", exists_ok=False)

    @classmethod
    def run(cls) -> None:
        if len(sys.argv) != 3:
            raise RuntimeError(f"Usage: python -m {cls.__module__} <task_key> <config_path>")
        task_key, config_path = sys.argv[1:]
        task = cls.from_components(task_key, Path(config_path), use_cli=False)

        if not isinstance(task, RunnableMixin):
            raise RuntimeError(f"Task {task} must be a `RunnableMixin`")

        # Adding the "running" lock file before rmoving the "scheduled" lock
        # file in order to prevent accidentally launching another job while the
        # current job is being set up.
        task.add_lock_file("running", exists_ok=True)
        task.remove_lock_file("scheduled", missing_ok=True)

        # Sets environment variables from Slurm environment variables.
        set_slurm_master_addr_and_port()
        rank, world_size = set_slurm_rank_and_world_size()

        # Sets the initialization method and configures per-rank logging.
        set_init_method("env://")
        configure_logging(rank=rank, world_size=world_size)
        init_process_group_from_backend()

        # Gets parallelism environment variables.
        model_parallelism = int(os.environ.get("MODEL_PARALLELISM", "1"))
        pipeline_parallelism = int(os.environ.get("PIPELINE_PARALLELISM", "1"))
        backend = os.environ.get("BACKEND", None)
        model_parallel_backend = os.environ.get("MODEL_PARALLEL_BACKEND", None)
        pipeline_parallel_backend = os.environ.get("PIPELINE_PARALLEL_BACKEND", None)
        data_parallel_backend = os.environ.get("DATA_PARALLEL_BACKEND", None)

        # Sets model parallelism.
        init_parallelism(
            model_parallelism=model_parallelism,
            pipeline_parallelism=pipeline_parallelism,
            mp_backend=model_parallel_backend or backend,
            pp_backend=pipeline_parallel_backend or backend,
            dp_backend=data_parallel_backend or backend,
        )

        task.add_signal_handler(requeue_job, signal.SIGUSR1)

        # Runs the base training loop.
        task.run()


if __name__ == "__main__":
    SlurmLauncher.run()
